# -*- perl -*-
#
# Copyright (C) 1996 Ken MacLeod
# See the file COPYING for distribution terms.
#

use strict;

use Text::GDO::Context;

package Text::GDO::toHTML;

#
# `initted' is true after the first attempt to run `ToHTML', when
# `ToHTML' calls `init'
#
$Text::GDO::toHTML::initted = 0;

sub _init {
    my ($prefix, $sdatadir, $filename);

    foreach $prefix (@::INC) {
	$sdatadir = "$prefix/sdata";
	if (-d $sdatadir) {
	    opendir (SDATADIR, $sdatadir)
		|| die "$::prog: Can't open directory \`$sdatadir' for reading: $!\n";
	    foreach $filename (readdir (SDATADIR)) {
		next if ($filename !~ /\.2ab$/);
		open (ENTITYFILE, "$sdatadir/$filename")
		    || die "$::prog: Can't open \`$sdatadir/$filename' for reading: $!\n";
		while (<ENTITYFILE>) {
		    chop;
		    s/(^[^\t]+)\t(.*)/$Text::GDO::toHTML::entities{"$1"} = $2/e;
		}
		close (ENTITYFILE);
	    }
	    closedir (SDATADIR);
	}
    }
    my (@rest) =
	(
	 'Screen',             "<PRE>",        "</PRE>\n",
	 'BlockQuote',         "<BLOCKQUOTE>", "</BLOCKQUOTE>\n",
	 'Line',               "",             "<BR>\n",
	 'ListTerm',           "<DT>",         "</DT>\n",

	 'Inline_Emphasis',    "<EM>",         "</EM>",
	 'Inline_Replaceable', "<I>",          "</I>",
	 'Inline_Literal',     "<SAMP>",         "</SAMP>",
	 );
    my ($key, $pre, $post);
    while (($key, $pre, $post, @rest) = @rest) {
	$Text::GDO::toHTML::pre{$key} = $pre;
	$Text::GDO::toHTML::post{$key} = $post;
    }

    # XXX fix in sdata
    $Text::GDO::toHTML::entities{"[bull  ]"} = "*";
    $Text::GDO::toHTML::entities{"[tilde ]"} = "~";

    # XXX LinuxDoc, should be fixed in the linuxdoc-dtd patch
    $Text::GDO::toHTML::entities{"</"} = "&lt;/";

    $Text::GDO::toHTML::initted = 1;
}

sub stdout {
    my ($self) = @_;

    print $self->{'data'};

    return 1;
}

sub write {
    my ($self, $basename) = @_;

    if (!open (OUTPUT, ">$basename.html")) {
	warn "$::prog: could not open $basename.html for writing, $!\n";
	return 0;
    }
    print OUTPUT $self->html_wrap ($basename, $self->{'data'});
    close (OUTPUT);

    my ($section);
    foreach $section (keys %{$self->{'sections'}}) {
	# XXX `html_add_content' shouldn't add a value under
        # 'sections' if we're >= max_depth, i.e. this `next' shouldn't
	# be necessary
	next if (!defined $self->{'sections'}{$section}{'data'});

	if (!open (OUTPUT, ">$basename-$section.html")) {
	    warn "$::prog: could not open $basename-$section.html for writing, $!\n";
	    return 0;
	}
	print OUTPUT $self->html_wrap ($basename, $self->{'sections'}{$section}{'data'});
	close (OUTPUT);
    }

    return 1;
}

sub html_wrap {
    my ($self, $basename, $data) = @_;
    my ($title, $sect_num);

    # perl re note: .*? means _shortest_ match
    $data =~ s|\01(.*?)\01|
	$sect_num = $1;
	$title = $self->{'sections'}{$sect_num}{'title'};
        # remove any embedded anchors
        $title =~ s;</?A[^>]*>;;g;
	qq{<A HREF="$basename-$sect_num.html">} . $title . qq{</A><BR>\n}
    |ge;
    $data =~ s|\02(.*?)\02|
	$sect_num = $1;
        qq{$basename-$sect_num.html}
    |ge;

    return $data;
}

sub scalar {
    my ($self) = @_;

    return $self->{'data'};
}

package Text::GDO::DocObject;

sub ToHTML {
    my ($self) = @_;

    my ($html_instance) = {};
    bless $html_instance, 'Text::GDO::toHTML';

    Text::GDO::toHTML::_init() if (!$Text::GDO::toHTML::initted);
    my ($context) = new Text::GDO::Context::HTML (
				     'entities' => \%Text::GDO::toHTML::entities,
				     'pre' => \%Text::GDO::toHTML::pre,
				     'post' => \%Text::GDO::toHTML::post,
				     'pref_compact' => 0,
				     'pref_block' => 'normal',
				     'rootnode' => $self->rootnode,
				     'html_instance' => $html_instance,
				     'max_depth' => 0,
				     'current' => [{
					 'fill' => '0',
					 'par_mode' => 'par',
					 'quoting' => 1,
					 'depth' => 0,
				     }]
				     );

    $html_instance->{'data'} = $self->html_to ($context);
    return $html_instance;
}

sub html_to {
    my ($self, $context) = @_;
    my ($thing, @inlines, $data);

    $context->push;
    $self->html_context ($context);
    $data = "";
    foreach $thing (@{$self->{'content'}}) {
	if ($thing->is_inline) {
	    push (@inlines, $thing);
	} else {
	    if ($#inlines != -1) {
		$data .= $self->html_format ($context, @inlines);
		@inlines = ();
	    }
	    $data .= $thing->html_to($context);
	}
    }
    if ($#inlines != -1) {
	$data .= $self->html_format ($context, @inlines);
    }
    $data = $self->html_add_content ($context, $data);
    $context->pop;

    return $data;
}

sub html_context {
}

sub html_add_content {
    my ($self, $context, $data) = @_;

    my ($ref);
    ($ref = ref $self) =~ s/^Text::GDO:://;

    return $context->{'pre'}{$ref} . $data . $context->{'post'}{$ref};
}

sub html_format {
    my ($self, $context, @inlines) = @_;
    my ($inline);
    my ($pre, $post, @data);

    my ($par_mode) = $context->par_mode;
    ($par_mode eq 'par')  && do { $pre = "<P>";  $post = "</P>\n" };
    ($par_mode eq 'brk')  && do { $pre = "";     $post = "<BR>\n" };
    ($par_mode eq 'none') && do { $pre = "";     $post = ""       };

    foreach $inline (@inlines) {
	push (@data, $inline->html_string ($context));
    }


    my ($data) = join ("", @data);
    if ($data =~ /^\s*$/s) {
	# This handles the case where a newline or whitespace creeps
	# in between two block-like elements while in an element that
	# accepts "PCDATA"
	return;
    } else {
	return $pre . $data . $post;
    }
}

sub html_escape {
    my ($self, $str) = @_;

    $str =~ s/&/&amp;/g;
    $str =~ s/</&lt;/g;
    $str =~ s/>/&gt;/g;

    return $str;
}

sub html_string {
    my ($self, $context) = @_;

    my ($ref);
    ($ref = ref $self) =~ s/^Text::GDO:://;
    my ($data) = $self->html_escape ($self->as_string ($context));

    return $context->{'pre'}{$ref} . $data . $context->{'post'}{$ref};
}

#
# Creates an endpoint as a header
# This will be the closest header element to the source document's
# `ID=' element
#
sub html_anchor {
    my ($self, $context, $level, $str) = @_;
    my ($id);

    if (defined ($self->{'id'})) {
	$id = $self->{'id'};
	if ($context->netscape || $context->html3) {
	    $str = "\n<H$level ID=\"$id\">"
		. $str . "</H$level>\n";
	} else {
	    # remove any embedded anchors
	    $str =~ s|</?A[^>]*>||g;
	    # The HTML2 DTD says A goes inside H
	    $str = "\n<H$level><A NAME=\"$id\">" . $str
	        . "</A></H$level>\n";
	}
    } else {
	$str = "\n<H$level>" . $str . "</H$level>\n";
    }

    return $str;
}

package Text::GDO::Inline_Quote;

sub html_string {
    my ($self, $context) = @_;
    my ($pre, $post);

    my ($data) = $self->Text::GDO::Inline::html_string ($context);

    if ($context->quoting) {
	$pre = $context->charent('[ldquo ]');
	$post = $context->charent('[rdquo ]');
    }

    return join ("", $pre, $data, $post);
}

package Text::GDO::Inline_Literal;

sub html_string {
    my ($self, $context) = @_;
    my ($pre, $post);

    my ($data) = $self->Text::GDO::Inline::html_string ($context);

    if ($context->quoting) {
	$pre = $context->charent('[lsquo ]');
	$post = $context->charent('[rsquo ]');
    }

    return join ("", $pre, $data, $post);
}

package Text::GDO::Author;

sub html_to {
    my ($self, $context) = @_;
    my (@data);

    $context->push ('par_mode' => 'brk');

    my ($align) = $context->align;

    my ($name) = $self->name;
    push (@data, "<B>$name</B><BR>\n")
	if ($name);

    push (@data, $self->{'title'}->html_to ($context))
	if (defined $self->{'title'});

    push (@data, $self->{'orgname'}->html_to ($context))
	if (defined $self->{'orgname'});

    my ($orgunit);
    foreach $orgunit (@{$self->{'orgunit'}}) {
	push (@data, $orgunit->html_to ($context));
    }

    my ($address) = $self->address;
    if ($address) {
	$address =~ s/\n/<BR>\n/s;
	push (@data, $address);
    }

    if (defined $self->{'email'}) {
	my ($email) = $self->{'email'}->as_string ($context);
	push (@data, "<A HREF='mailto:$email'>&lt;$email&gt;</A><BR>\n");
    }

    $context->pop;

    return "<P$align>" . join ("", @data) . "</P>\n";
}

package Text::GDO::Document;

sub html_add_content {
    my ($self, $context, $data) = @_;
    my (@data);

    my ($align) = $context->set_align ($context->{'netscape'} ? " ALIGN=CENTER" : "");

    $context->push ('par_mode' => 'none');
    push (@data, "<H1$align>" . $self->{'title'}->html_to ($context) . "</H1>\n")
	if (defined $self->{'title'});
    push (@data, "<H2$align>" . $self->{'sub_title'}->html_to ($context) . "</H2>\n")
	if (defined $self->{'sub_title'});

    push (@data, "<H4$align>" . $self->{'date'}->html_to ($context) . "</H4>\n")
	if (defined $self->{'date'});
    $context->pop;

    my ($author);
    foreach $author (@{$self->{'authors'}}) {
	push (@data, $author->html_to ($context));
    }

    if (defined $self->{'abstract'}) {
	push (@data, "<H4$align><I>ABSTRACT</I></H4>\n");
	push (@data, "<BLOCKQUOTE>\n"
	      . $self->{'abstract'}->html_to ($context) . "</BLOCKQUOTE>\n");
    }

    push (@data, $self->{'copyright'}->html_to ($context))
	if (defined $self->{'copyright'});

    return join ("", @data, $data);
}

package Text::GDO::Section;

sub html_to {
    my ($self, $context) = @_;

    if ($context->depth >= $context->max_depth) {
	return $self->Text::GDO::DocObject::html_to ($context);
    } else {
	my ($html_section) = {};

	my ($sect_num) = $self->formal_number;
	if ($sect_num eq "") {
	    die "toHTML: assert: sect_num has no formal_number\n";
	}

	$context->{'html_instance'}{'sections'}{$sect_num} = $html_section;

	$html_section->{'data'} = $self->Text::GDO::DocObject::html_to ($context);
	$html_section->{'sect_num'} = $sect_num;
	$html_section->{'doc_object'} = $self;

	return "\1$sect_num\1";
    }
}

sub html_add_content {
    my ($self, $context, $data) = @_;

    my ($sect_num_raw) = $self->formal_number;
    my ($sect_num) = $sect_num_raw;
    if ($sect_num ne "") {
	# Numbered section
	$sect_num = $self->formal_number . "  ";
    }
    my ($sect_level) = $self->level;

    my ($title);
    if (defined $self->{'title'}) {
	$context->push ('par_mode' => 'none');
	$title = $sect_num . $self->{'title'}->html_to ($context);
	$context->pop;
    } else {
	$title = $sect_num . "";
    }

    if ($sect_num_raw ne "") {  # XXX asserted in html_to, but...
	$context->{'html_instance'}{'sections'}{$sect_num_raw}{'title'} = $title;
    }

    if ($sect_level > 6) {$sect_level = 6};

    return $self->html_anchor ($context, $sect_level, $title) . $data;
}

package Text::GDO::Formal;

sub html_add_content {
    my ($self, $context, $data) = @_;
    my (@data);

    if (defined $self->{'title'}) {
	$context->push ('par_mode' => 'none');
	push (@data,
	      $self->html_anchor ($context, 4, $self->formal_name . "  "
				  . $self->{'title'}->html_to ($context)));
	$context->pop;
    }

    return join ("", @data, $data);
}

package Text::GDO::Screen;

sub html_context {
    my ($self, $context) = @_;

    $context->set_par_mode ('none');
    $context->set_quoting (0);
}

package Text::GDO::NoFill;
#
# `NoFill' in HTML must be like a `Screen' to work
#
sub html_context {
    my ($self, $context) = @_;

    $context->set_par_mode ('none');
}

package Text::GDO::Admonition;

sub html_add_content {
    my ($self, $context, $data) = @_;
    my (@data);

    if (defined $self->{'title'}) {
	$context->push ('par_mode' => 'none');
	push (@data,
	      $self->html_anchor ($context, 4,
				  $self->{'title'}->html_to ($context)));
	$context->pop;
    } else {
	push (@data, $self->html_anchor ($context, 4, $self->type));
    }

    return join ("", @data, $data);
}

package Text::GDO::List;

sub html_context {
    my ($self, $context) = @_;

    my ($list_type) = $context->set_list_type ($self->type);
}

sub html_add_content {
    my ($self, $context, $data) = @_;
    my ($pre, $post) = ("<UL>\n", "</UL>\n");

    my ($list_type) = $context->list_type;
  SWITCH:
    {
	($list_type eq 'ordered')
	    && do { $pre = "<OL>\n";  $post = "</OL>\n"; last SWITCH };
	($list_type eq 'itemized')
	    && do { $pre = "<UL>\n";  $post = "</UL>\n"; last SWITCH };
	($list_type eq 'variable')
	    && do { $pre = "<DL>\n";  $post = "</DL>\n"; last SWITCH };
    }

    return join ("", $pre, $data, $post);
}

package Text::GDO::ListItem;

sub html_add_content {
    my ($self, $context, $data) = @_;
    my ($pre, $post) = ("<LI>", "</LI>\n");

    # XXX this is probably browser specific
    # remove paragraph tags around first paragraph and after last
    if ($data =~ /^\s*<P>/) {
	$data =~ s/^\s*<P>//s;
	$data =~ s|</P>||s; # note, no 'g', only first occurrence
    }
    $data =~ s|</P>\s*$||s;

    my ($list_type) = $context->list_type;
  SWITCH:
    {
	($list_type eq 'ordered')
	    && do { $pre = "<LI>";  $post = "</LI>\n"; last SWITCH };
	($list_type eq 'itemized')
	    && do { $pre = "<LI>";  $post = "</LI>\n"; last SWITCH };
	($list_type eq 'variable')
	    && do { $pre = "<DD>";  $post = "</DD>\n"; last SWITCH };
    }

    return join ("", $pre, $data, $post);
}

package Text::GDO::ListTerm;

sub html_context {
    my ($self, $context) = @_;

    $context->set_par_mode ('none');
}

package Text::GDO::Table;

sub html_to {
    my ($self, $context) = @_;

    # XXX implement html3/netscape tables

    my ($data) = $self->html_escape ($self->ToAscii()->scalar);

    return join ("", "<PRE>", $data, "</PRE>\n");
}

package Text::GDO::XRefEnd;

sub html_string {
    my ($self, $context) = @_;
    my ($linkend, $id_attr, $title_obj);
    my ($title);

    $title_obj = $context->{'rootnode'}{'referables'}{$self->{'link'}}{'object'};
#XXX   $title_obj = $context->reference($self->{'link'});
#XXX sigh.  wish I knew how to do "if (defined &$title_obj->formal_name())"
    if (defined $title_obj && eval { $title = $title_obj->formal_name() }) {
	my ($title_num) = $title_obj->formal_number();
	my ($self_num) = $self->Text::GDO::Section::formal_number();
	my ($title_doc) = "";
	if ($title_num ne $self_num && $context->max_depth != 0) {
	    $title_doc = "\2$title_num\2";
	}
	$title = $title . ", <I>"
	    . $title_obj->{'title'}->html_string ($context) . "</I>";
	# remove any embedded anchors
	$title =~ s|</?A[^>]*>||g;
	$title = "<A HREF=\"$title_doc#" . $self->{'link'} . "\">"
	    . $title . "</A>";
	return $title;
    } else {
#XXX in the case of LinuxDoc's LABEL, the DocObject is known, but we
#don't know where it is
	return "[[Unknown link reference]]";
    }
}

package Text::GDO::URL;

sub html_string {
    my ($self, $context) = @_;
    # XXX urls and names should be undef not ""
    my ($url) = "";

    if (defined $self->{'url'}) {
	$url = $self->{'url'}->as_string (Text::GDO::Context::string_context());
    }
    my ($name) = $self->Text::GDO::DocObject::ascii_string($context);
    if ($name ne "" && $name ne $url) {
	return "<A HREF=\"$url\">$name</A>";
    } elsif ($url eq "") {
	return "<A HREF=\"$name\">$name</A>";
    } else {
	return "<A HREF=\"$url\">$url</A>";
    }
}

package Text::GDO::SGML::Event;

sub html_string {
    my ($event, $context) = @_;

    return Text::GDO::DocObject::html_escape ($event, $event->as_string ($context));
}

package Text::GDO::Context::HTML;
@Text::GDO::Context::HTML::ISA = qw{Text::GDO::Context};

sub max_depth     { return $_[0]->{'max_depth'} }
sub netscape      { return $_[0]->{'netscape'} }
sub html3         { return $_[0]->{'html3'} }

sub max_depth     { return $_[0]->{'max_depth'} = $_[1] }
sub set_netscape  { return $_[0]->{'netscape'} = $_[1] }
sub set_html3     { return $_[0]->{'html3'} = $_[1] }

# inherits `fill'
sub align         { return $_[0]->{'current'}[-1]{'align'} }
sub depth         { return $_[0]->{'current'}[-1]{'depth'} }
sub par_mode      { return $_[0]->{'current'}[-1]{'par_mode'} }
sub list_type     { return $_[0]->{'current'}[-1]{'list_type'} }
sub quoting       { return $_[0]->{'current'}[-1]{'quoting'} }

sub set_align     { return $_[0]->{'current'}[-1]{'align'} = $_[1] }
sub depth         { return $_[0]->{'current'}[-1]{'depth'} = $_[1] }
sub set_par_mode  { return $_[0]->{'current'}[-1]{'par_mode'} = $_[1] }
sub set_list_type { return $_[0]->{'current'}[-1]{'list_type'} = $_[1] }
sub set_quoting   { return $_[0]->{'current'}[-1]{'quoting'} = $_[1] }

sub push_sections { return push (@{$_[0]->{'current'}{'sections'}}, $_[1]) }
sub pop_sections  { return pop (@{$_[0]->{'current'}{'sections'}}) }
sub append_sections { return push (@{$_[0]->{'current'}{'sections'}[-1]}, $_[1]) }

1;
